/*
CDBeQuiet! - Slows down your optical drive under DOS
Copyright (C) 2017  Christoph Boelitz

E-mail: c.boelitz@web.de
Project page: http://cdbq.dosforum.de/

CDBeQuiet! is fully free of charge and may be
distributed. Its source code may be modified as long
as you leave my contact information intact. Using it
in any commercial manner is prohibited.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!

*/

//Includes
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <dos.h>

#include "ide.h"
#include "log.h"
#include "main.h"

//Global vars
PORT prm_port = 0;
int prm_force = 0;
int prm_device = 0;
int prm_log = 0;
int prm_reset = 0;
int prm_info = 0;
int prm_help = 0;
int prm_speed = -1;

void exit_func()
{
  //Called on exit

  if (prm_log)
    close_logfile();

  puts("");
}

char* trim(char *str)
{
    //Trims a string in memory
    //Returns new pointer to the same string

    int len;
	int ptr;
	
	len = strlen(str);;

     //First trim leading spaces

    while (str[0] == ' ')
      str++;

    //Trim trailing spaces

    ptr = len-1;
	
    while (str[ptr] == ' ')
      ptr--;

    str[ptr+1] = '\0';

    return str;
}

void cmdline_help()
{
  //Displays help screen for command-line options

  const char *cmd_help =
    "Usage: cdbq -p port -d device -s speed [-irfl]\n"
    "\n"
    "Command-line options:\n\n"
    "  -p port    Specify primary (port = 1), secondary (port = 2),\n"
    "               tertiary (port = 3), quarternary (port = 4).\n"
    "  -d device  Specify master (device = m) or slave (device = s) device.\n"
    "  -s speed   Choose a custom CD read speed in kb/s.\n"
    "               (affects spindle speed); 0 for max speed\n"
    "  -i         Just print out device information and exit.\n"
    "  -r         Perform soft reset of CD-ROM drive and exit (experimental).\n"
    "  -f         Skip all sanity checks and force sending command.\n"
    "  -l         Log screen output to file cdbq.log (implies /v).\n"
    "  -h         This help screen\n"
    "\n"
    "Example:\n"
    "  If you wanted to set a CD speed of 4x (1x = 150 kb/s)\n"
    "  for a master CD-ROM on your secondary IDE channel,\n"
    "  the following command will do:\n"
    "    cdbq -p2 -dm -s600"
  ;

  puts(cmd_help);
}

void handle_error(int err_code, int log, ...)
{
  //Error handling and output

  va_list args;

  va_start(args, log);

  switch(err_code)
  {
    case IDE_ERR_TIME:
    case LOG_ERR_TIME:
      textout_log(1, log, "\n!! ERROR: Cannot measure time.\n\n");
    break;
    case LOG_ERR_LOGFILE_OPEN:
      textout_log(1, log, "\n!! ERROR: Cannot open log file.\n\n");
    break;
    case IDE_ERR_COMMAND_ABORTED:
      textout_log(1, log, "\n!! ERROR: ATA command aborted unexpectedly.\n\n");
    break;
    case IDE_ERR_ATAPI_SRST_ABORTED:
      textout_log(1, log, "\n!! ERROR: ATAPI soft reset command was aborted unexpectedly.\n"
                            "!!        Maybe the device doesn't support this command..\n\n");
    break;
    case IDE_ERR_IDDRIVE_NOT_ABORTED:
      textout_log(1, log, "\n!! ERROR: IDENTIFY DRIVE was not aborted as expected.\n"
                            "!!        Please check that device is not a hard disk.\n\n");
    break;
    case IDE_ERR_WAITBUSY:
      textout_log(1, log, "\n!! ERROR: Timeout waiting for IDE to clear busy status.\n\n");
    break;
    case IDE_ERR_WAITDRQ:
      textout_log(1, log, "\n!! ERROR: Timeout waiting for high DRQ.\n\n");
    break;
    case IDE_ERR_WAITDRDY:
      textout_log(1, log, "\n!! ERROR: Device not ready.\n\n");
    break;
    case IDE_ERR_NOT_ATAPI_DEV:
      textout_log(1, log, "\n!! ERROR: Device does not speak ATAPI.\n\n");
    break;
    case IDE_ERR_NO_ATAPI_SIG:
      textout_log(1, log, "\n!! ERROR: No ATAPI signature found.\n\n");
    break;
    case IDE_ERR_NOT_A_CDROM:
      textout_log(1, log, "\n!! ERROR: Device speaks ATAPI, but is not a CD-ROM.\n\n");
    break;
    case IDE_ERR_WRONG_PKT_SIZE:
      textout_log(1, log, "\n!! ERROR: Device has wrong packet size.\n\n");
    break;
    case CDBQ_ERR_CMDLINE_UNK_PRINT:
      textout_log(1, log,
        "\n!! ERROR: Unknown command-line option `-%c'.\n"
          "!!        Check syntax for errors or missing params using -h.\n\n",
        args
      );
    break;
    case CDBQ_ERR_CMDLINE_UNK_NONPRINT:
      textout_log(1, log,
        "\n!! ERROR: Unknown command-line option character '\\x%x'.\n"
          "!!        Check syntax for errors or missing params using -h.\n\n",
        args
      );
    break;
    case CDBQ_ERR_CMDLINE_MISSING_ARG:
    {
      int c;

      c = va_arg(args, int);
      textout_log(1, log,
        "\n!! ERROR: Missing command-line argument for option -%c.\n"
        "!!        Check syntax for errors or missing params using -h.\n\n",
        c
      );
      break;
    }
    case CDBQ_ERR_CMDLINE_MISSING_OPT:
    {
      int c;

      c = va_arg(args, int);
      textout_log(1, log,
        "\n!! ERROR: Missing command-line option.\n"
        "!!        -p, -d and -s are mandatory options. Check help using -h.\n\n",
        c
      );
      break;
    }
    case CDBQ_ERR_CMDLINE_GETOPT:
      textout_log(1, log,
        "\n!! ERROR: Could not parse command-line.\n"
          "!!        Unexpected getopt() behavior. Check syntax using -h.\n\n");
    break;
    case CDBQ_ERR_CMDLINE_CONFLICT:
      textout_log(1, log, "\n!! ERROR: Conflicting command-line options detected.\n\n");
    break;
    case CDBQ_ERR_CMDLINE_WRONG_SPEED:
      textout_log(1, log, "\n!! ERROR: Speed parameter must be between 0 and 4096 kb/s.\n\n");
    break;
    default:
      textout_log(1, log,
        "\n!! ERROR: Unknown error. Code: %d\n", err_code);

  }

  va_end(args);

  //Critical error, terminate program
  exit(err_code);
}




void cmdline_parse(int argc, char *argv[])
{
  //Handling of command line arguments
  int c;
  
  opterr = 0;

  while ((c = getopt (argc, argv, "irlfp:d:s:h")) != -1)
  {
    switch (c)
    {
      //Switches
    case 'i':
      prm_info = 1;
    break;
    case 'r':
      prm_reset = 1;
    break;
    case 'l':
      prm_log = 1;
    break;
    case 'f':
      prm_force = 1;
    break;
    case 'h':
      prm_help = 1;
    break;
    //Parameters that take an argument
    case 'p':
      if (strcmp(optarg, "1") == 0)
        prm_port = 1; //IDE1_BASE
      else if (strcmp(optarg, "2") == 0)
        prm_port = 2; //IDE2_BASE
      else
        prm_port = atoi(optarg);
    break;

    case 'd':
      if (strcmp(optarg, "m") == 0)
        prm_device = 1;
      else if (strcmp(optarg, "s") == 0)
        prm_device = 2;
    break;

    case 's':
      prm_speed = atoi(optarg);
    break;

    //getopt() error occured
    case '?':
      if (optopt == 'p' || optopt == 'd' || optopt == 's')
        handle_error(CDBQ_ERR_CMDLINE_MISSING_ARG, 0, optopt);
      else if (isprint (optopt))
        handle_error(CDBQ_ERR_CMDLINE_UNK_PRINT, 0, optopt);
      else
        handle_error(CDBQ_ERR_CMDLINE_UNK_NONPRINT, 0, optopt);
    break;
    default:
      handle_error(CDBQ_ERR_CMDLINE_GETOPT, 0);
    }
  }

  /*
  //Debug
  printf(
      "prm_info: %d\n"
      "prm_help: %d\n"
      "prm_reset: %d\n"
      "prm_log: %d\n"
      "prm_force: %d\n"
      "prm_port: %d\n"
      "prm_device: %d\n"
      "prm_speed: %d\n",
      prm_info, prm_help, prm_reset, prm_log,
      prm_force, prm_port, prm_device, prm_speed
  );
  */

  //Check if all mandatory options have been given

  if
  (
    (!prm_port || (prm_speed == -1 && !prm_reset && !prm_info) ||
    !prm_device) && !prm_help
  )
    handle_error(CDBQ_ERR_CMDLINE_MISSING_OPT, 0);

  //Conflicting options selected as in reset and speed
  if
  (
    (prm_info && prm_reset)        ||
    (prm_info && prm_speed != -1)  ||
    (prm_info && prm_force)        ||
    (prm_reset && prm_speed != -1)
  )
    handle_error(CDBQ_ERR_CMDLINE_CONFLICT, 0);

  //CD speed not correct
  if (prm_speed != -1 && (prm_speed < 0 || prm_speed > 4096))
    handle_error(CDBQ_ERR_CMDLINE_WRONG_SPEED, 0);
}



int main(int argc, char *argv[])
{
  //Main program entry point

  int ret; //Used for various return values
  PORT ide_port; //Port parameter
  int ide_device; //Master/slave parameter
  char ide_str[40]; //e.g. "primary master"
  drive_info di;
  WORD packet[6]; //ATAPI packet command

  atexit(exit_func);

  cmdline_parse(argc, argv);

  //Logfile
  if (prm_log)
  {
     ret = open_logfile(CDBQ_LOGFILE);

     if (ret != LOG_ERR_SUCCESS)
       handle_error(ret, 0); //Logging errors obviously can't be logged
  }

  textout_log(0, prm_log,
    "\n"
    "CDBeQuiet! v" CDBQ_VER "\n"
    "  by Christoph Blitz (c.boelitz@web.de)\n"
  );

  if (prm_help)
  {
    cmdline_help();
    return 0;
  }

  //Port parameter
  if (prm_port == 1) ide_port = IDE1_BASE;
  else if(prm_port == 2) ide_port = IDE2_BASE;
  else if(prm_port == 3) ide_port = IDE3_BASE;
  else if(prm_port == 4) ide_port = IDE4_BASE;
  else ide_port = prm_port;

  //Master/slave parameter
  if (prm_device == 1) ide_device = IDE_DEVICE_MASTER;
  else ide_device = IDE_DEVICE_SLAVE;

  if (prm_port == 1) sprintf(ide_str, "primary");
  else if (prm_port == 2) sprintf(ide_str, "secondary");
  else if (prm_port == 3) sprintf(ide_str, "tertiary");
  else if (prm_port == 4) sprintf(ide_str, "quarternary");
  else sprintf(ide_str, "port 0x%x", prm_port);

  if (prm_device == 1) strcat(ide_str, " master");
  else strcat(ide_str, " slave");

  //Check device information

  if (prm_force && !prm_info) goto skip_checks;

  textout_log(0, prm_log, "\nReading drive information from %s device...\n", ide_str);

  ret = ide_get_atapi_info(ide_port, ide_device, &di);

  if (ret != IDE_ERR_SUCCESS)
    handle_error(ret, prm_log);

  textout_log(0, prm_log, "  Model number  : %s\n", trim(di.model_number));
  textout_log(0, prm_log, "  Firmware rev. : %s\n", trim(di.firmware_revision));

  //Now some precautions...

  ret = ide_check_if_cdrom(&di);

  if (ret != IDE_ERR_SUCCESS)
    handle_error(ret, prm_log);

  if (prm_info) //We're done here.
    return 0;

  skip_checks:

  if (prm_reset)
  {
    //Issue soft reset command and quit
    textout_log(0, prm_log, "\nSending software reset to %s device...\n", ide_str);

    ret = ide_atapi_soft_reset(ide_port, ide_device);

    if (ret != IDE_ERR_SUCCESS)
      handle_error(ret, prm_log);

    textout_log(0, prm_log, "\nDevice is resetting. It can take up to 30 secs until it's ready again.\n", ide_str);

    //We're done here.
    return 0;
  }

  if (prm_speed == 0)
    textout_log(0, prm_log, "\nSetting maximum CD-ROM read rate...\n", ide_str);
  else
    textout_log(0, prm_log, "\nSetting CD-ROM read rate of %d kb/s on %s device...\n", prm_speed, ide_str);

  //Send ATAPI command packet
  ide_atapi_packet_set_cd_speed(packet, prm_speed);

  ret = ide_atapi_send_packet_cmd(ide_port, ide_device, packet);

  if (ret != IDE_ERR_SUCCESS)
    handle_error(ret, prm_log);

  if (prm_speed == 0)
    textout_log(0, prm_log, "\nNew speed set.\n"); //Maximum speed, no silence :)
  else
    textout_log(0, prm_log, "\nNew speed set. Enjoy the silence...\n");

  return 0;
}
